/* Materials used:
	1. VIP Help (First and foremost).
	2. Ivan Bratko. Prolog (Programming for Artificial Intellegence).
	3. VIP examples (A_136_2.PRO, A_146_2.PRO, geni.pro).
	4. Mark Minasi. The Complete PC Upgrade and Maintainance Guide.
	5. Giarratano & Riley. Expert Systems (Principles and programming).
*/
	
domains
	CondNum=Integer.
	RuleNum=Integer.
	Conditions=CondNum*.
	HistoryTree=RuleNum*.
	Kind=String.
	
facts
	rule(RuleNum, Kind, Kind, Conditions).
	condition(CondNum, String).
	yes(CondNum).
	no(CondNum).

predicates
%	Inference engine
	nondeterm examine(HistoryTree, Kind).
	check(RuleNum, HistoryTree, Conditions).
	noTest(CondNum).

%	Getting data from user. askUser is a hull, processAnswer is a core part.
	askUser(HistoryTree, RuleNum, CondNum, String).
	processAnswer(HistoryTree, RuleNum, String, CondNum, Integer).

%	Providing explanatory data for user.
	subCat(Kind, Kind, Kind).
	showConditions(Conditions, String).
	showRule(RuleNum, String).
	report(HistoryTree, String).
	reverse(Conditions, Conditions).
	reverse1(Conditions, Conditions, Conditions).

clauses
%	Main predicate. Invokes routine-action perfoming predicates.
	examine(_, MyGoal) :-
	  not(rule(_, MyGoal, _, _)),
	  !,
	  nl,
	  write("I think that you should ", MyGoal),
	  !.

	examine(HistoryTree, MyGoal) :-
	  rule(RuleNum, MyGoal, SubGoal, Conditions),
	  check(RuleNum, HistoryTree, Conditions),
	  examine([RuleNum|HistoryTree], SubGoal).

%	Checking conditions before implementing a rule
	check(RuleNum, HistoryTree, [CondListHead|CondListTail]) :-
	  yes(CondListHead), % User answered positively
	  !,
	  check(RuleNum, HistoryTree, CondListTail). % Continue 

	check(_, _, [CondListHead|_]) :-
	  no(CondListHead), % User answered negatively. Stop further checking.
	  !,
	  fail.

%	Two predicates dealing with conditions starting with 'not'
	check(RuleNum, HistoryTree, [CondListHead|CondListTail]) :-
	  condition(CondListHead, ConditionString),
	  fronttoken(ConditionString, "not", _NewCondString), % Delete not from ConditionString if suitable
	  frontchar(_NewCondString ,_, NextCondString), % Delete one character from _NewCondString
	  condition(NewCondListHead, NextCondString),
	  noTest(NewCondListHead),
	  !,
	  check(RuleNum, HistoryTree, CondListTail).
	
%	See above
	check(_, _, [CondListHead|_]) :-
	  condition(CondListHead, ConditionString),
	  fronttoken(ConditionString, "not", _NewCondString),
	  frontchar(_NewCondString, _, NextCondString),
	  condition(NewCondListHead, NextCondString),
	  yes(NewCondListHead),
	  !,
	  fail.

%	Call askUser predicate if above mentioned don't work
	check(RuleNum, History, [CondListHead|CondListTail]) :-
	  condition(CondListHead, Text),
	  !,
	  askUser(History, RuleNum, CondListHead, Text),
	  check(RuleNum, History, CondListTail).

	check(_,_,[]).

%	Helping predicates for negatively formulated conditions
	noTest(Condition) :-
	  no(Condition),
	  !.
	
	noTest(Condition) :-
	  not(yes(Condition)),
	  !.

	askUser(HistoryTree, RuleNum, CondNum, Text) :-
	  write("Is the following statement true?\n", Text, "\n(Yes/No/Why): "),
	  readchar(Key), % Get input from user
	  upper_lower(Key, Answer), % Convert it to lowercase
	  processAnswer(HistoryTree, RuleNum, Text, CondNum, Answer).

	processAnswer(_, _, _, CondNum,'y') :-
	  !,
	  assert(yes(CondNum)), % Add to knowledge base
	  write("Yes\n").
  
	processAnswer(_, _, _, CondNum, 'n') :-
	  !,
	  assert(no(CondNum)),
	  write("No\n"),
	  fail.

%	Answering user 'why' questions and displaying them good formated
	processAnswer(HistoryTree, RuleNum, Text, CondNum, 'w') :-
	  !,
	  write("Why\n"),
	  rule(RuleNum, StartGoal, SubGoal, _),
	  !,
	  subCat(StartGoal, Subgoal, String1),
	  concat("I am trying to prove that ", String1, String2),
	  concat(String2, "\nBy using rule number ", String3),
	  str_int(RuleNumAsString, RuleNum), % Convert integer to string.
	  concat(String3, RuleNumAsString, Answer1),
	  showRule(RuleNum, String4),
	  concat(Answer1, String4, Answer2),
	  report(HistoryTree, Output),
	  concat(Answer2, Output, Answer3),
	  write(Answer3),
	  nl,
	  askUser(HistoryTree, RuleNum, CondNum, Text).

%	Otherwise
	processAnswer(HistoryTree, RuleNum, Text, CondNum, _) :-
	  write("\nDidn't understand. Please, again.\n"),
	  askUser(HistoryTree, RuleNum, CondNum, Text).
	
	showRule(RuleNum, String) :-
	  rule(RuleNum, MyGoal1, MyGoal2, CondList),
	  !,
	  str_int(RuleNumAsString, RuleNum), % Convert RuleNum from integer to string
	  concat("\n Rule ", RuleNumAsString, Answer), % Following four lines just to make output nicer
	  concat(Answer, ": ", Answer1),
	  subCat(MyGoal1, MyGoal2, String1),
	  concat(Answer1, String1, Answer2),
	  concat(Answer2, "\n     if ", Answer3),
	  reverse(CondList, ReversedCondList), % Output to user in better way
	  showConditions(ReversedCondList, Conditions),
	  concat(Answer3, Conditions, String).
 
	showConditions([], ""). % Nothing to show
  
	showConditions([Conditions], Answer) :-
	  condition(Conditions, Answer),
	  !.
  
	showConditions([Head|Tail], Answer) :-
	  condition(Head, Text),
	  !,
	  concat("\n          and ", Text, NextString),
	  showConditions(Tail, NextAnswer),
	  concat(NextAnswer, NextString, Answer).

%	Provides good indenting
	subCat(MyGoal1, MyGoal2, String) :-
	  format(String, "If % then %", MyGoal1, MyGoal2).

	report([], ""). % Nothing to report

	report([RuleListHead|RuleListTail], String) :-
	rule(RuleListHead, MyGoal1, MyGoal2,_),
	!,
	subCat(MyGoal1, MyGoal2, String1),
	concat("\nI have shown that: ", String1, String2),
	concat(String2,"\nBy using rule number ", String3),
	str_int(RuleListHeadAsString, RuleListHead),
	concat(String3, RuleListHeadAsString, String4),
	concat(String4, ":\n ", String5),
	showRule(RuleListHead, Str),
	concat(String5, Str, String6),
	report(RuleListTail, NextString),
	concat(String6, NextString, String).
	
	reverse(Conditions, NewConditions) :-
	  reverse1([], Conditions, NewConditions).

	reverse1(Conditions, [], Conditions).

	reverse1(Conditions, [CondListHead|CondListTail], NewConditions) :-
	  reverse1([CondListHead|Conditions], CondListTail, NewConditions).
	
goal
	consult("sage.db"),
	write("Have a computer problem? Let me help you."),
	nl, nl,
	examine([], "there is electricity in area");
	write("\nStrange. Ok, I surrender, what's the correct answer?!"),
	nl, nl.